﻿B4A=true
Group=Default Group
ModulesStructureVersion=1
Type=Class
Version=9.85
@EndOfDesignText@
'Class Module: B4XMainPage
'Project: DomoticzQuickViewer
'Select a Domoticz plan and view the devices in a custom listview.
'A plan selected can be a room plan or other like Favorites, All Devices, All Used Devices, All Not Used Devices. 
'References:
'https://www.b4x.com/android/forum/threads/b4x-xui-views-cross-platform-views-and-dialogs.100836/#content
'List of page related events in the B4XPagesManager object. The event name is B4XPage.
'20220203 rwbl

#Region Notes
' Android.jar: ensure(!) to set minimum android-28\android.jar (else error unable to find attribute android:fontVariationSettings)
' Android App Design Guidelines: https://developer.android.com/design/
' Domoticz JSON/API: https://www.domoticz.com/wiki/Domoticz_API/JSON_URL's
#End Region


#Region Shared Files
#CustomBuildAction: folders ready, %WINDIR%\System32\Robocopy.exe,"..\..\Shared Files" "..\Files"
'Ctrl + click to sync files: ide://run?file=%WINDIR%\System32\Robocopy.exe&args=..\..\Shared+Files&args=..\Files&FilesSync=True
#End Region

'Ctrl + click to export as zip: ide://run?File=%B4X%\Zipper.jar&Args=Project.zip

Sub Class_Globals
	'Version & Info
	Private TITLE As String			= "Domoticz Quick Viewer"
	Private VERSION As String 		= "v1.00 (Build 20220203)"
	Private ABOUT As String			= $"${TITLE}${CRLF}${VERSION}${CRLF}${CRLF}Copyright 2022 (C) by Robert W.B. Linn (https://github.com/rwbl)${CRLF}${CRLF}GNU General Public License Version 3 (see LICENSE)."$

	'Debug
	'Private mDebug As Boolean = True
	Private mDebug As Boolean = False
	
	'UI
	Private Root As B4XView
	#If B4J
	Private fx As JFX
	#End If
	Private xui As XUI
	'Toolbar & Bottom bar
	Private cbPlans As B4XComboBox
	Private btnSearch As B4XView
	Private liServer As B4XLoadingIndicator
	Private btnRefreshPlans As B4XView
	Private btnRefresh As B4XView
	Private lblStatusMessage As B4XView
	'CLV with label items from layout clvDevicesitem
	Private clvDevices As CustomListView
	Private lblDeviceName As B4XView
	Private lblDeviceData As B4XView
	Private lblDeviceLastUpdate As B4XView
	Private ivDeviceGroup As B4XImageView
	Private btnSetDevice As B4XView
	'Dialogs
	Private deviceDialog As B4XDialog
	'Preferences - The json file in the files folder defines the preference dialog structure with items type, key, title, required fields
	Private prefDialog As PreferencesDialog
	Private prefOptions As Map
	Private prefOptionsFile As String 				= "preferences.map"
	Private PREF_DIALOG_FILE As String 				= "preferences.json"		'Definition of the preferences. Ensure to lowercase the preference key (see files folder)
	Private DomoticzProductionIP As String 			= DOMOTICZ_PRODUCTION_IP	'"key": "domoticzproductionip"
	Private DomoticzDevelopmentIP As String 		= DOMOTICZ_DEVELOPMENT_IP	'"key": "domoticzdevelopmentip"
	Private UseDomoticzProduction As String			= False						'"key": "usedomoticzproduction"
	Private RefreshInterval As Int 					= 0							'"key": "refreshinterval"
	'Menu items - accessible via burger or 3 vertical dots icon on the titlebar
	Public MENU_PREFERENCES As String 				= "Preferences"	'
	Public MENU_ABOUT As String						= "About"
	'Domoticz Controller
	Private DomController As DomoticzController
	Private DOMOTICZ_PRODUCTION_IP As String 		= "NNN.NNN.NNN.NNN"
	Private DOMOTICZ_DEVELOPMENT_IP As String		= "NNN.NNN.NNN.NNN"
	'Refresh timer
	Private Timer_Refresh As Timer
	Private btnPreferences As B4XView
	'CLV Devices
	Private CLV_ITEM_HEIGHT As Int 					= 66dip
	Private CLV_ITEM_TEXTSIZE As Double				= 14
	Private itemScrollTo As Boolean					= False
	Private itemSelected As Int						= -1
End Sub

Public Sub Initialize
	Log($"${TITLE} ${VERSION}"$)
'	B4XPages.GetManager.LogEvents = True
End Sub

Sub MainForm_Resize (Width As Double, Height As Double)
	If deviceDialog.IsInitialized Then
		If deviceDialog.Visible Then deviceDialog.Resize(Width, Height)
	End If
End Sub

'Don't miss the code in the Main module + manifest editor.
Private Sub IME_HeightChanged (NewHeight As Int, OldHeight As Int)
	prefDialog.KeyboardHeightChanged(NewHeight)
End Sub

#Region B4XPAGE
'This event will be called once, before the page becomes visible.
Private Sub B4XPage_Created (Root1 As B4XView)
	'Root and B4XPages
	Root = Root1
	Root.LoadLayout("mainpage")
	B4XPages.SetTitle(Me, TITLE)
	'UI views Toolbar (LoadIndicator, Plan selection set focus), CustomListView, Font sizes
	liServer.Hide
	' clvDevices.DefaultTextBackgroundColor = xui.Color_Black
	cbPlans.cmbBox.RequestFocus
	'Dialog - The title is set in the selected dialog
	deviceDialog.Initialize(Root)
	deviceDialog.Title = ""
	'Init the Domotizc Controller and the timer. These are not enabled as done in InitPreferences > LoadPreferences
	DomController.Initialize("", 0, mDebug)
	Timer_Refresh.Initialize("Timer_Refresh", 0)
	Timer_Refresh.Enabled = False
	'Init the Preferences and take actions accordingly
	InitPreferences
	'Get the plans and the device data
	CallSubDelayed(Me, "Action_Refresh_Plans")
End Sub

Private Sub B4XPage_CloseRequest As ResumableSub
	If xui.IsB4A Then
		'Handle back key in Android
		If prefDialog.BackKeyPressed Then Return False
	End If
	Return True
End Sub
#End Region

#Region REFRESHTIMER
'Handle refresh timer tick events. The interval is set in the preferences in minutes.
Sub Timer_Refresh_Tick
	'Check the index
	If cbPlans.SelectedIndex < 0 Then Return
	Action_Refresh_Plan
End Sub
#End Region

#Region TOOLBAR
'Views from left to right
'Handle plans dropdown selection.
'The selected list index is used to get the plan map. From the plan map, the idx is used to get the plan devices with properties.
Private Sub cbPlans_SelectedIndexChanged(Index As Int)
	'Check the index
	If Index < 0 Then Return
	'Get the plan map from the list of plans
	Dim Plan As Map = DomController.Plans.Get(Index)
	'Get the plan idx from the plan map
	Dim PlanIdx As Int = Plan.Get(DomController.DOMOTICZ_KEY_IDX)
	ToLog($"cbPlans_SelectedIndexChanged: index=${Index}, idx=${PlanIdx}"$)
	'Get the devices for the given plan idx
	Action_GetDevices(PlanIdx)
End Sub

Private Sub btnSearch_Click
	Action_Search_Item
End Sub

Private Sub btnRefresh_Click
	Action_Refresh_Plan
End Sub

Private Sub btnRefreshPlans_Click
	Action_Refresh_Plans
End Sub

Private Sub btnPreferences_Click
	Action_Dialog_Preferences	
End Sub
#End Region

#Region ACTIONS
'Get the list of Domoticz devices depending planid.
'The customlistview devices is updated if there are devices found. No action is taken if the flag statussuccess is false.
Private Sub Action_GetDevices(PlanID As Int)
	'Show the loading indicator in the toolbar
	liServer.Show
	'Get the list of devices from the Domoticz server
	Wait For(DomController.GetDevices(PlanID)) Complete (ResultState As Boolean)
	'Wait For(DomController.GetStatusFavoriteDevices) Complete (ResultState As Boolean)
	ToLog($"Action_GetDevices: planid=${PlanID}, statussuccess=${DomController.StatusSuccess}, statusmessage=${DomController.StatusMessage}"$)
	If DomController.StatusSuccess Then
		'Clear all items first from the clv
		clvDevices.Clear
		For Each Device As Map In DomController.Devices
			'Add the device selected properties with device idx as value
			clvDevices.Add(clvDevices_ItemCreate(Device, clvDevices.AsView.Width, CLV_ITEM_HEIGHT), Device)
		Next
		If itemScrollTo Then
			CallSubDelayed2(Me, "Action_ScrollToItem", itemSelected)
		End If
		lblStatusMessage.Text = $"${DomController.ServerIP}, #Devices found: ${DomController.Devices.Size}, Last update: ${DateTime.Time(DateTime.Now)}"$
	End If
	'Hide the loading indicator in the toolbar
	liServer.Hide
End Sub

'Scroll to an item in the clv.
Private Sub Action_ScrollToItem(Index As Int)
	If clvDevices.Size > 0 And Index <> -1 And Index < clvDevices.Size - 1 Then
		clvDevices.ScrollToItem(Index)
	End If
	itemScrollTo = False
End Sub

'Refresh the list of devices for the plan selected in the dropdown.
Private Sub Action_Refresh_Plan
	If cbPlans.SelectedIndex < 0 Then Return
	cbPlans_SelectedIndexChanged(cbPlans.SelectedIndex)
End Sub

'Get the list of plans, i.e. custom and room plans. The first plan entry "Favorites" is selected.
Private Sub Action_Refresh_Plans
	liServer.Show
	'Get the list of rooms from the Domoticz server
	Wait For(DomController.GetPlans) Complete (ResultState As Boolean)
	ToLog($"Action_Refresh_Plans: statussuccess=${DomController.StatusSuccess}, statusmessage=${DomController.StatusMessage}"$)
	If DomController.StatusSuccess Then
		Dim Plans As List
		Plans.Initialize
		For Each Plan As Map In DomController.Plans
			Plans.Add(Plan.Get(DomController.DOMOTICZ_KEY_NAME))
		Next
		cbPlans.SetItems(Plans)
		If cbPlans.Size > 0 Then
			cbPlans_SelectedIndexChanged(0)
		End If
	End If
	liServer.Hide
End Sub

Private Sub Action_Select_Plan(idx As String)
	liServer.Show
	'Clear all items first from the clv
	clvDevices.Clear
	'Get the list of devices from the Domoticz server
	Wait For(DomController.GetPlanDevices(idx)) Complete (ResultState As Boolean)
	'Wait For(DomController.GetStatusFavoriteDevices) Complete (ResultState As Boolean)
	ToLog($"Action_Select_Plan: statussuccess=${DomController.StatusSuccess}, statusmessage=${DomController.StatusMessage}"$)
	If DomController.StatusSuccess Then
		For Each Device As Map In DomController.Devices
			'Add the device selected properties with device idx as value
			clvDevices.Add(clvDevices_ItemCreate(Device, clvDevices.AsView.Width, CLV_ITEM_HEIGHT), Device)
		Next
	End If
	liServer.Hide
End Sub

'Search for an item by selecting its name from the selection list.
Private Sub Action_Search_Item
	Dim SearchTemplate As B4XSearchTemplate
	SearchTemplate.Initialize
	Dim DeviceNames As List
	DeviceNames = DomController.GetDeviceNamesList
	SearchTemplate.SetItems(DeviceNames)
	Dim rs As Object = deviceDialog.ShowTemplate(SearchTemplate, "OK", "", "Cancel")
	Sleep(100)
	Wait For (rs) Complete (Result As Int)
	ToLog($"Action_Search_Item: result=${Result}, selecteditem=${SearchTemplate.SelectedItem}"$)
	If Result = xui.DialogResponse_Positive Then
		Dim Index As Int = DeviceNames.IndexOf(SearchTemplate.SelectedItem)
		clvDevices.JumpToItem(Index)
		ToLog($"Action_Search_Item: selecteditem=${SearchTemplate.SelectedItem}, index=${Index}"$)
	End If
End Sub

'Show the Domoticz device log for the selected device.
'The browser will be opened. This can be slow on Android.
'B4A: Requires the library Phone
'ShowDeviceLog(Device.Get(DomController.DOMOTICZ_KEY_IDX))
Private Sub Action_Device_Log(Idx As Int)	'ignore
	Dim url As String = DomController.GetDeviceLogUrl(Idx)
	ToLog($"ShowDeviceLog: idx=${Idx}"$)
	#If B4A
	Dim pi As PhoneIntents
	StartActivity(pi.OpenBrowser(url))
	#End If
	#If B4J
	fx.ShowExternalDocument(url)
	#End If
End Sub

'Handle the button device value setting from the selected clv item.
Private Sub Action_Device_Set(Index As Int)
	' Dim Index As Int = clvDevices.GetItemFromView(Sender)
	Dim device As Map = DomController.Devices.Get(Index)
	Dim TITLE As String
	Dim Label As String
	Dim Value As String
	Dim UserVariableName, UserVariableType As String
	ToLog($"btnSetDevice_Click name=${device.Get(DomController.DOMOTICZ_KEY_NAME)}, data=${device.Get(DomController.DOMOTICZ_KEY_DATA)}, dialogid=${DomController.GetDialogID(device)}"$)
	'Depending device type/subtype open a b4xdialog
	DomController.StatusSuccess = False
	itemSelected = Index
	Select DomController.GetDialogID(device)

		Case DomController.DIALOGID_SWITCHLIGHT
			TITLE = $"Device: ${device.Get(DomController.DOMOTICZ_KEY_NAME)}"$
			Value = device.Get(DomController.DOMOTICZ_KEY_DATA).As(String).ToUpperCase
			Label = $"Switch state: ${Value}"$
			Wait For(DomController.Dialog_SwitchLight(deviceDialog, TITLE, Label, Value)) Complete (SwitchLightResult As Boolean)
			If DomController.StatusSuccess Then
				'Set the switch depending ResultValue True=On, or False=Off
				Dim SwitchCmd As String = IIf(SwitchLightResult, "On", "Off")
				Wait For(DomController.SetSwitchLight(device.Get(DomController.DOMOTICZ_KEY_IDX), SwitchCmd)) Complete (ResultValue As Boolean)
			End If

		Case DomController.DIALOGID_SWITCHLIGHT_SETLEVEL
			Dim Level As Int = device.Get(DomController.DOMOTICZ_KEY_LEVEL)
			Dim LevelNames As String = device.Get(DomController.DOMOTICZ_KEY_LEVELNAMES)
			TITLE = $"${device.Get(DomController.DOMOTICZ_KEY_NAME)} - ${DomController.GetLevelName(Level, LevelNames)}"$
			Value = device.Get(DomController.DOMOTICZ_KEY_LEVEL)
			Label = $"Selector state: ${Value}"$
			Wait For(DomController.Dialog_SwitchLightSetLevel(deviceDialog, TITLE, LevelNames, Value)) Complete (SwitchLightSetLevelResult As Int)
			If DomController.StatusSuccess Then
				'Set the switch depending ResultValue True=On, or False=Off
				Wait For(DomController.SetSwitchLightSetLevel(device.Get(DomController.DOMOTICZ_KEY_IDX), SwitchLightSetLevelResult)) Complete (ResultValue As Boolean)
			End If

		Case DomController.DIALOGID_SETSETPOINT
			TITLE = $"Device: ${device.Get(DomController.DOMOTICZ_KEY_NAME)}"$
			Value = device.Get(DomController.DOMOTICZ_KEY_DATA).As(Float)
			Label = $"Setpoint (Current: ${Value}):"$
			Wait For(DomController.Dialog_Positive(deviceDialog, TITLE, Label, Value)) Complete (SetSetpointResult As Float)
			If DomController.StatusSuccess Then
				Wait For(DomController.SetSetpoint(device.Get(DomController.DOMOTICZ_KEY_IDX), SetSetpointResult)) Complete (ResultValue As Boolean)
			End If
			
		Case DomController.DIALOGID_USERVARIABLE
			TITLE = $"User Variable: ${device.Get(DomController.DOMOTICZ_KEY_NAME)}"$
			Value = device.Get(DomController.DOMOTICZ_KEY_DATA)
			Label = $"Value: ${Value}"$
			UserVariableName = device.Get(DomController.DOMOTICZ_KEY_NAME)
			UserVariableType = device.Get(DomController.DOMOTICZ_KEY_TYPE)
			Wait For(DomController.Dialog_Text(deviceDialog, TITLE, Label, Value)) Complete (UserVariableResult As String)
			If DomController.StatusSuccess Then				
				Wait For(DomController.SetUserVariable(UserVariableName, UserVariableType, UserVariableResult)) Complete (ResultValue As Boolean)
			End If			

		Case Else
			xui.MsgboxAsync($"[ERROR] Can not set device."$, $"Set Device ${device.Get(DomController.DOMOTICZ_KEY_NAME)}"$)
	End Select
	'Refresh the list
	If DomController.StatusSuccess Then
		itemScrollTo = True
		CallSubDelayed(Me, "Action_Refresh_Plan")
	End If
End Sub

'Show the preference dialog.
'Note the setting of the field explanation "appversion" with global string "ABOUT"
'If OK, preferences are saved, the local preference global vars are updated, the plan list refreshed and the device data loaded.
Private Sub Action_Dialog_Preferences
	prefDialog.SetExplanation("about", $"${ABOUT}"$)
	Wait For (prefDialog.ShowDialog(prefOptions, "OK", "CANCEL")) Complete (Result As Int)
	If Result = xui.DialogResponse_Positive Then
		SavePreferences
		Sleep(100)
		'Get the plans and the device data - required because the server ip might have changed, i.e. production <> development
		CallSubDelayed(Me, "Action_Refresh_Plans")
	End If
End Sub
#End Region

#Region CLV
'Create a CLV item.
'Device - Map with all device properties
'Width - Width of the pane
'Height - Height of the pane
#If B4A
Sub clvDevices_ItemCreate(Device As Map, Width As Int, Height As Int) As Panel
#End If
#If B4J
Sub clvDevices_ItemCreate(Device As Map, Width As Int, Height As Int) As Pane
#End If
	Dim p As B4XView = xui.CreatePanel("")
	p.SetLayoutAnimated(0, 0, 0, Width, Height)
	p.LoadLayout("clvdevicesitem")
	' Log($"${Device.Get(DomController.DOMOTICZ_KEY_NAME)}, ${Device.Get(DomController.DOMOTICZ_KEY_TYPE)}, ${Device.Get(DomController.DOMOTICZ_KEY_SUBTYPE)}"$)
	'ivDeviceGroup.Bitmap = FontAwesomeToBitmap(DomController.GetDeviceGroupIcon(Device.Get(DomController.DOMOTICZ_KEY_TYPE), Device.Get(DomController.DOMOTICZ_KEY_SUBTYPE)), 16, xui.Color_Black)
	lblDeviceName.Text = Device.Get(DomController.DOMOTICZ_KEY_NAME)
	lblDeviceName.TextSize = CLV_ITEM_TEXTSIZE
	lblDeviceData.Text = Device.Get(DomController.DOMOTICZ_KEY_DATA)
	lblDeviceData.TextSize = CLV_ITEM_TEXTSIZE
	lblDeviceLastUpdate.Text = Device.Get(DomController.DOMOTICZ_KEY_LASTUPDATE)
	lblDeviceLastUpdate.TextSize = CLV_ITEM_TEXTSIZE
	'Special cases
	'Selector add selected level name as text
	If Device.Get(DomController.DOMOTICZ_KEY_LEVELNAMES) <> Null Then
		Dim Level As Int = Device.Get(DomController.DOMOTICZ_KEY_LEVEL)
		Dim LevelNames As String = Device.Get(DomController.DOMOTICZ_KEY_LEVELNAMES)
		lblDeviceData.Text = $"${lblDeviceData.Text } (${DomController.GetLevelName(Level, LevelNames)})"$
	End If
	'Show a button to set device value if the dialogid is not information - NOT ALL DEVICE TYPES ARE SUPPORTED
	btnSetDevice.Visible = IIf(DomController.GetDialogID(Device) == DomController.DialogID_INFORMATION, False, True)
	ToLog($"${Device.Get(DomController.DOMOTICZ_KEY_NAME)} ${Device.Get(DomController.DOMOTICZ_KEY_IDX)}=${Device.Get(DomController.DOMOTICZ_KEY_TYPE)} ${Device.Get(DomController.DOMOTICZ_KEY_SUBTYPE)}"$)
	Return p
End Sub

'Handle customlistview devices item click.
'In a messagebox, selected device properties are shown with the option to show the device Log in a browser.
'TODO: Add more properties as required.
Private Sub clvDevices_ItemClick (Index As Int, Value As Object)
	ToLog($"clvDevices_ItemClick: index=${Index}, value=${Value}"$)
	Dim Device As Map = Value
	ToLog($"clvDevices_ItemClick: device=${DomController.MapToString(Device)}"$)
	itemSelected = Index
	Dim sb As StringBuilder
	sb.Initialize
	sb.Append($"Name: ${Device.Get(DomController.DOMOTICZ_KEY_NAME)}"$).Append(CRLF)
	sb.Append(CRLF)
	sb.Append($"IDX: ${Device.Get(DomController.DOMOTICZ_KEY_IDX)}"$).Append(CRLF)
	sb.Append($"Description: ${Device.Get(DomController.DOMOTICZ_KEY_DESCRIPTION)}"$).Append(CRLF)
	sb.Append(CRLF)
	sb.Append($"Data: ${Device.Get(DomController.DOMOTICZ_KEY_DATA)}"$).Append(CRLF)
	sb.Append(CRLF)
	If Device.Get(DomController.DOMOTICZ_KEY_SUBTYPE) == DomController.DOMOTICZ_VALUE_SUBTYPE_USERVARIABLE Then
		sb.Append($"Type: ${Device.Get(DomController.DOMOTICZ_KEY_TYPE)} (${DomController.GetUserVariableType(Device.Get(DomController.DOMOTICZ_KEY_TYPE))})"$).Append(CRLF)
	Else
		sb.Append($"Type: ${Device.Get(DomController.DOMOTICZ_KEY_TYPE)}"$).Append(CRLF)
	End If
	sb.Append($"SubType: ${Device.Get(DomController.DOMOTICZ_KEY_SUBTYPE)}"$).Append(CRLF)
	If Device.Get(DomController.DOMOTICZ_KEY_LEVELNAMES) <> Null Then
		sb.Append($"LevelNames: ${DomController.Base64Decode(Device.Get(DomController.DOMOTICZ_KEY_LEVELNAMES))}"$).Append(CRLF)
	End If
	sb.Append(CRLF)
	sb.Append($"Last Update: ${Device.Get(DomController.DOMOTICZ_KEY_LASTUPDATE)}"$).Append(CRLF)
	'xui.MsgboxAsync($"${sb.ToString}"$, $"Device Information"$)
	'Dialog showing selected device properties with the option to show the device Log.
	Dim sf As Object = xui.Msgbox2Async($"${sb.ToString}"$, $"Device Information"$, "Close", "Log", "", Null)
	Wait For (sf) Msgbox_Result (Result As Int)
	If Result = xui.DialogResponse_Cancel Then
		Action_Device_Log(Device.Get(DomController.DOMOTICZ_KEY_IDX))
	End If
End Sub

'Handle the button device value setting from the selected clv item.
Private Sub btnSetDevice_Click
	Dim Index As Int = clvDevices.GetItemFromView(Sender)
	Action_Device_Set(Index)
End Sub
#End Region

#Region SETTINGS
'Init the preferences loading from json file located in the assets folder (same as the layout files etc.)
'After prefdialog has been initialized, load the preference data from the map file into the preferences options map used to populate the preference dialog fields.
Private Sub InitPreferences
	prefOptions.Initialize
	xui.SetDataFolder ("preferences")
	prefDialog.Initialize(Root, "Preferences", 300dip, 400dip)
	prefDialog.Title = "Preferences"
	prefDialog.LoadFromJson(File.ReadString(File.DirAssets, PREF_DIALOG_FILE))
	'Set the input event listener which call the event PrefDialog_IsValid
	prefDialog.SetEventsListener(Me, "PrefDialog")
	LoadPreferences
End Sub

'Verify preferences
Private Sub PrefDialog_IsValid (TempData As Map) As Boolean
	'Check if the refresh interval is not < 0. If 0 the refreshinterval is disabled.
	Dim interval As Int = TempData.GetDefault("refreshinterval", 0)
	If interval < 0 Then
		prefDialog.ScrollToItemWithError("refreshinterval")
		Return False
	End If
	Return True
End Sub

'Load the preferences from a file holding the prefOptions Map object
'Android Device: Folder=/data/data/make.rwbl.boostrccar/files, File=boostrccar.map
Private Sub LoadPreferences
	'Set defaults first
	prefOptions.Put("domoticzproductionip", $"${DOMOTICZ_PRODUCTION_IP}"$)
	prefOptions.Put("domoticzdevelopmentip", $"${DOMOTICZ_DEVELOPMENT_IP}"$)
	prefOptions.Put("usedomoticzproduction", UseDomoticzProduction)
	prefOptions.Put("refreshinterval", RefreshInterval)
	'If there is a preference file, get the preferences from there
	If File.Exists(xui.DefaultFolder, prefOptionsFile) Then
		Try
			prefOptions = File.ReadMap(xui.DefaultFolder, prefOptionsFile)
		Catch
			xui.MsgboxAsync($"[ERROR] Can not load the preferences.${CRLF}${LastException}"$, "Load Preferences")
			Log($"[ERROR] LoadPreferences - File: ${xui.DefaultFolder}, ${prefOptionsFile}, ${LastException}"$)
		End Try
	End If
	'Populate the global vars from the prefOptions map
	DomoticzProductionIP	= prefOptions.Get("domoticzproductionip")
	DomoticzDevelopmentIP	= prefOptions.Get("domoticzdevelopmentip")
	UseDomoticzProduction	= prefOptions.get("usedomoticzproduction")
	RefreshInterval			= prefOptions.Get("refreshinterval")
	'Update the Domoticz Server IP and the Refresh Timer
	DomController.SetServerIP(IIf(UseDomoticzProduction, DomoticzProductionIP, DomoticzDevelopmentIP))
	'Update the refresh timer. The interval is converted from minutes to milliseconds - 0 = disable timer.
	Timer_Refresh.Interval = RefreshInterval * 60 * 1000
	Timer_Refresh.Enabled = IIf(RefreshInterval > 0, True, False)
	ToLog($"LoadPreferences - OK: ${DomController.MapToString(prefOptions)}"$)
End Sub

'Save the preferences map to the prefoptions file located in the xui default folder (see sub initpreferences).
'Reload the preferences to the local vars.
Private Sub SavePreferences
	Try
		'Checks prior saving the preferences to the file
		If prefOptions.Get("domoticzproductionip").As(String).Length == 0 Then prefOptions.Put(DOMOTICZ_PRODUCTION_IP, 0)
		'Write the preference map to the preference file
		File.WriteMap(xui.DefaultFolder, prefOptionsFile, prefOptions)
		ToLog($"SavePreferences - OK: folder=${xui.DefaultFolder}, file=${prefOptionsFile}"$)
		'Reload the preferences
		CallSubDelayed(Me, "LoadPreferences")
	Catch
		xui.MsgboxAsync($"[ERROR] Can not save the preferences.${CRLF}${LastException}"$, "Save Preferences")
		Log($"[ERROR] SavePreferences: ${LastException}"$)
	End Try
End Sub
#End Region

#Region HELPERS
'Write a message to the log with time as prefix text.
Private Sub ToLog(msg As String)
	If mDebug Then
		Log($"${DateTime.Time(DateTime.Now)} ${msg}"$)
	End If
End Sub

'Create a b4xbitmap 32x32 from a fontawesome icon
'Example: FontAwesomeToBitmap(Chr(0xF021), 26, xui.Color_White)
Private Sub FontAwesomeToBitmap(Text As String, FontSize As Float, Color As Int) As B4XBitmap	'ignore
	Dim xui As XUI
	#If B4A
	Dim p As Panel = xui.CreatePanel("")
	#End If
	#If B4J
	Dim p As Pane = xui.CreatePanel("")
	#End If
	p.SetLayoutAnimated(0, 0, 0, 32dip, 32dip)

	Dim cvs1 As B4XCanvas
	cvs1.Initialize(p)

	#If B4A
	Dim fnt As B4XFont = xui.CreateFont(Typeface.FONTAWESOME, FontSize)
	#End If
	#If B4J
	Dim fnt As B4XFont = xui.CreateFont(fx.CreateFontAwesome(FontSize), FontSize)
	#End If
	Dim r As B4XRect = cvs1.MeasureText(Text, fnt)
	
	Dim BaseLine As Int = cvs1.TargetRect.CenterY - r.Height / 2 - r.Top
	cvs1.DrawText(Text, cvs1.TargetRect.CenterX, BaseLine, fnt, Color, "CENTER")
	Dim b As B4XBitmap = cvs1.CreateBitmap
	cvs1.Release
	Return b
End Sub
#End Region
