﻿B4A=true
Group=Default Group
ModulesStructureVersion=1
Type=Class
Version=11.2
@EndOfDesignText@
'Class Module: DomoticzController
'Domoticz controller functions.
'Getter and Setter
'Dialogs
'References:
'https://www.b4x.com/android/forum/threads/b4x-resumable-subs-that-return-values-resumablesub.82670/
'https://www.domoticz.com/wiki/Domoticz_API/JSON_URL%27s
'https://www.b4x.com/android/forum/threads/b4x-xui-views-cross-platform-views-And-dialogs.100836/#content
'https://www.b4x.com/android/forum/threads/b4x-input-dialogs-with-xui-views.101197/
'20220202 rwbl

#Region Class Attributes 
#End Region

Sub Class_Globals
	Public VERSION As String 							= "DomoticzController v20220202"
	
	'UI
	Private xui As XUI
	'Dummy view to avoid B4A message "Activity context is required"
	Private DummyButton As Button 'ignore
	
	' Domoticz Server
	Public SERVER_IP_PREFIX As String 					= "http://"
	Public SERVER_IP_SUFFIX As String 					= ":8080/json.htm?"
	Public SERVER_IP_DEFAULT As String					= "NNN.NNN.NNN.NNN"

	'URL JSON/API requests
	'Get all the devices, i.e. used + unused + hidden
	Private URL_GET_DEVICES As String					= $"http://#IP#:8080/json.htm?type=devices"$	'ignore
	Private URL_GET_DEVICES_USED As String				= $"http://#IP#:8080/json.htm?type=devices&used=true&filter=all&displayhidden=1"$	'ignore
	Private URL_GET_DEVICES_NOTUSED As String			= $"http://#IP#:8080/json.htm?type=devices&used=false"$	'ignore

	Private URL_GET_DEVICES_FAVORITE As String			= $"http://#IP#:8080/json.htm?type=devices&used=true&filter=all&favorite=1"$	'ignore
	Private URL_GET_DEVICE As String					= $"http://#IP#:8080/json.htm?type=devices&rid=#IDX#"$	'ignore
	Private URL_GET_DEVICES_LIST As String				= $"http://#IP#:8080/json.htm?type=command&param=devices_list"$	'ignore

	Private URL_GET_PLANS As String						= $"http://#IP#:8080/json.htm?type=plans&order=name&used=True"$	'ignore
	Private URL_GET_PLAN_DEVICES As String				= $"http://#IP#:8080/json.htm?type=devices&plan=#IDX#"$	'ignore
	'Private URL_GET_PLAN_DEVICES As String				= $"http://#IP#:8080/json.htm?type=command&param=getplandevices&idx=#IDX#"$	'ignore

	Private URL_GET_USERVARIABLES As String				= $"http://#IP#:8080/json.htm?type=command&param=getuservariables"$	'ignore
	Private URL_SET_USERVARIABLE As String				= $"http://#IP#:8080/json.htm?type=command&param=updateuservariable&vname=#USERVARIABLENAME#&vtype=#USERVARIABLETYPE#&vvalue=#USERVARIABLEVALUE#"$
	
	'Dedicated device commands - the param name is used for the constants name
	Private URL_SWITCHLIGHT As String					= $"http://#IP#:8080/json.htm?type=command&param=switchlight&idx=#IDX#&switchcmd=#SWITCHCMD#"$
	Private URL_SWITCHLIGHT_SETLEVEL As String			= $"http://#IP#:8080/json.htm?type=command&param=switchlight&idx=#IDX#&switchcmd=Set%20Level&level=#LEVEL#"$
	Private URL_SETSETPOINT As String					= $"http://#IP#:8080/json.htm?type=command&param=setsetpoint&idx=#IDX#&setpoint=#SETPOINT#"$

	'Log commands
	Private URL_LOG_DEVICE As String					= $"http://#IP#:8080/#/Devices/#IDX#/Log"$
	
	'Placeholder for URL JSON/API requests (allways uppercase!)
	Private PH_IP As String 							= "#IP#"
	Private PH_IDX As String							= "#IDX#"
	Private PH_SWITCHCMD As String						= "#SWITCHCMD#"
	Private PH_LEVEL As String							= "#LEVEL#"
	Private PH_SETPOINT As String						= "#SETPOINT#"
	Private PH_USERVARIABLENAME As String				= "#USERVARIABLENAME#"
	Private PH_USERVARIABLETYPE As String				= "#USERVARIABLETYPE#"
	Private PH_USERVARIABLEVALUE As String				= "#USERVARIABLEVALUE#"

	'PlanIDs - The Domoticz defined PlanID are > 0. The PlanID < 0 are custom plan IDs, also used for sorting the devices list by Name.
	Public PLANID_DEVICES_FAVORITE As Int				= -1
	Public PLANID_DEVICES_ALL As Int					= -2
	Public PLANID_DEVICES_USED As Int					= -3
	Public PLANID_DEVICES_NOTUSED As Int				= -4
	Public PLANID_USERVARIABLES As Int					= -5
	Public PLANID_ROOMS_SEPERATOR As Int				= -99

	'Additional pland defined as room()s added to the rooms list retrieved from Domoticz. These are used in the UI dropdown select plan.
	'These are the first entries followed by the Domoticz defined rooms. The plan id's are previous defined.
	'From the JSON structure (matching the Domoticz URL response /json.htm?type=plans&order=name&used=True) the Namer and idx are used.
	Private PLAN_DEVICES_FAVORITES As String			= $"{"Devices": 0, "Name" : "Favorites", "Order" : "1", "idx" : "${PLANID_DEVICES_FAVORITE}"}"$
	Private PLAN_DEVICES_ALL As String					= $"{"Devices": 0, "Name" : "All Devices", "Order" : "1", "idx" : "${PLANID_DEVICES_ALL}"}"$
	Private PLAN_DEVICES_USED As String					= $"{"Devices": 0, "Name" : "Used Devices", "Order" : "1", "idx" : "${PLANID_DEVICES_USED}"}"$	
	Private PLAN_DEVICES_NOTUSED As String				= $"{"Devices": 0, "Name" : "Not Used Devices", "Order" : "1", "idx" : "${PLANID_DEVICES_NOTUSED}"}"$
	Private PLAN_USERVARIABLES As String				= $"{"Devices": 0, "Name" : "User Variables", "Order" : "1", "idx" : "${PLANID_USERVARIABLES}"}"$
	Private PLAN_ROOMS_SEPERATOR As String				= $"{"Devices": 0, "Name" : "-- Rooms --", "Order" : "1", "idx" : "${PLANID_ROOMS_SEPERATOR}"}"$

	'Device Groups
	Private GROUP_WEATHER As String 					= "Weather"
	Private GROUP_SWITCHES As String 					= "Switches"
	Private GROUP_INFORMATION As String 				= "Information"
	Private GROUP_UTILITY As String 					= "Utility"
	Private GROUP_SECURITY As String					= "Security"
	Private GROUP_UNKNOWN As String 					= "Unknown"		'ignore
	Private GROUP_ICON_WEATHER As String 				= Chr(0xF185)
	Private GROUP_ICON_SWITCHES As String 				= Chr(0xF0EB)
	Private GROUP_ICON_INFORMATION As String 			= Chr(0xF1EA)
	Private GROUP_ICON_UTILITY As String 				= Chr(0xF0AD)
	Private GROUP_ICON_SECURITY As String				= Chr(0xF25A)
	Private GROUP_ICON_UNKNOWN As String 				= Chr(0xF071)	'ignore warning icon

	' Messages & Text
	Private MSG_STATUS_OK As String 					= "OK"
	Private MSG_STATUS_ERROR As String 					= "ERROR"
	Private MSG_APPLY_SETTINGS As String				= "Restart the app to apply the settings."	'ignore
	Private	MSG_DATA_REFRESHED As String				= "Data refreshed #DATE#"					'ignore

	'
	' JSON
	'
	' JSON keys used in the Domoticz REST request result
	' The key must exact match!
	' Important to check (from time to time) any changes by placing REST request via  browser.

	Public DOMOTICZ_KEY_APPVERSION As String 			= "app_version"

	Public DOMOTICZ_KEY_DATA As String 					= "Data"
	Public DOMOTICZ_KEY_DAYLENGTH As String				= "DayLength"
	Public DOMOTICZ_KEY_DESCRIPTION As String			= "Description"
	Public DOMOTICZ_KEY_FAVORITE As String 				= "Favorite"
	Public DOMOTICZ_KEY_GUST As String 					= "Gust"
	Public DOMOTICZ_KEY_IDX As String 					= "idx"
	Public DOMOTICZ_KEY_LASTUPDATE As String 			= "LastUpdate"
	Public DOMOTICZ_KEY_LEVEL As String 				= "Level"
	Public DOMOTICZ_KEY_LEVELNAMES As String			= "LevelNames"
	Public DOMOTICZ_KEY_NAME As String 					= "Name"
	Public DOMOTICZ_KEY_PLANID As String 				= "PlanID"
	Public DOMOTICZ_KEY_RESULT As String 				= "result"
	Public DOMOTICZ_KEY_SERVERTIME As String 			= "ServerTime"
	Public DOMOTICZ_KEY_STATUS As String 				= "Status"
	Public DOMOTICZ_KEY_SUBTYPE As String 				= "SubType"
	Public DOMOTICZ_KEY_SUNRISE As String 				= "Sunrise"
	Public DOMOTICZ_KEY_SUNSET As String 				= "Sunset"
	Public DOMOTICZ_KEY_TYPE As String 					= "Type"
	Public DOMOTICZ_KEY_USED As String 					= "Used"
	Public DOMOTICZ_KEY_VALUE As String 				= "Value"

	Public DOMOTICZ_KEY_DIRECTIONSTR As String			= "DirectionStr"
	Public DOMOTICZ_KEY_RETURN_STATUS As String 		= "status"
	Public DOMOTICZ_KEY_RETURN_TITLE As String 			= "title"

	Public DOMOTICZ_KEY_SWITCH_ON_VALUE   As String 	= "On"
	Public DOMOTICZ_KEY_SWITCH_OFF_VALUE  As String 	= "Off"
	Public DOMOTICZ_KEY_SWITCH_TYPE As String 			= "SwitchType"
	Public DOMOTICZ_KEY_SWITCH_TYPE_DIMMER As String	= "Dimmer"
	Public DOMOTICZ_KEY_SWITCH_TYPE_ONOFF As String 	= "On/Off"
	Public DOMOTICZ_KEY_SECURITY_NORMAL As String 		= "Normal"

	Public DOMOTICZ_KEY_TYPE_GENERAL As String 			= "General"
	Public DOMOTICZ_KEY_TYPE_WIND As String 			= "Wind"
	Public DOMOTICZ_KEY_TYPE_SECURITY As String 		= "Security"

	Public DOMOTICZ_VALUE_SUBTYPE_USERVARIABLE As String 	= "UserVariable"	'Added to handle user variables as devices

	'The devices with planid < 0, i.e. the custom plans are sorted by field (Domoticz key) "Name"
	Private DEVICE_SORT_FIELDNAME As String				= DOMOTICZ_KEY_NAME

	'Defaults for settings or other
	Public WIND_UNIT_DEFAULT As String 					= "bf"
	Public MESSAGE_DEVICE_DEFAULT As Int 				= 0

	'Dialog IDs
	Public DIALOGID_UNKNOWN  As Int 					= -1
	Public DialogID_INFORMATION  As Int 				= 1
	Public DIALOGID_INPUT  As Int 						= 2
	Public DIALOGID_SWITCHLIGHT As Int 					= 3
	Public DIALOGID_SWITCHLIGHT_SETLEVEL As Int 		= 4
	Public DIALOGID_SETSETPOINT As Int 					= 5
	Public DIALOGID_USERVARIABLE As Int 				= 20
	
	'Public objects
	Type TDevice (Idx As Int, Name As String)
	Public Devices As List
	Public Plans As List
	Public UserVariables As List

	'Status flag and message
	Public StatusSuccess As Boolean
	Public StatusMessage As String

	'Class locals (start with lowercase m prefix)
	Private mDebug As Boolean
	Private mServerIP As String
	Private mRefreshInterval As Int	'ignore
	Private mDeviceGroups As List
	Private mRooms As List
	Private mUserVariabelTypes() As String = Array As String("Integer", "Float", "String", "Date", "Time")

'	Private mDeviceProperties() As String = Array As String(DOMOTICZ_KEY_IDX, DOMOTICZ_KEY_NAME, DOMOTICZ_KEY_DESCRIPTION, DOMOTICZ_KEY_TYPE, DOMOTICZ_KEY_SUBTYPE, DOMOTICZ_KEY_DATA)

'	Private mJSONResponseRoot As Map
'	Private mJSONResponseResult As List
End Sub

'Initializes the object.
'ServerIP - IP address of the domoticz server
'RefreshInterval - Domoticz data refresh interval (0=disabled)
'Debug - Flag to enable detailed logging
Public Sub Initialize(ServerIP As String, RefreshInterval As Int, Debug As Boolean)
	If ServerIP.Length == 0 Then ServerIP = SERVER_IP_DEFAULT
	mServerIP = ServerIP
	mRefreshInterval = RefreshInterval
	mDeviceGroups = SetDeviceGroups
	mRooms.Initialize
	mDebug = Debug
	Devices.Initialize
	Plans.Initialize
	UserVariables.Initialize
End Sub

#Region GETTER
'Get the status for devices depending the given planid. 
'The class global Devicesis updated and can be used as required to get device(s) selected properties.
'PlanID - Favorites, devices used, devices notused, uservariables or plans from the roomslist.
'Returns - Boolean True=OK, False=Error. Check class globals StatusSuccess and StatusMessage.
'Example: <code>
'Wait For(DomController.GetDevices(PlanID)) Complete (ResultState As Boolean)
'If DomController.StatusSuccess Then
'	clvDevices.Clear
'	For Each Device As Map In DomController.Devices
'		'Add the device selected properties with device idx as value
'		clvDevices.Add(clvDevices_ItemCreate(Device, clvDevices.AsView.Width, CLV_ITEM_HEIGHT), Device)
'	Next
'End If
'</code>
Public Sub GetDevices(PlanID As String) As ResumableSub
	Dim Result As Boolean = False
	Dim Job As HttpJob
	Dim Url As String
	'Set the url depending planid
	Select PlanID
		Case PLANID_DEVICES_FAVORITE
			Url = URL_GET_DEVICES_FAVORITE.Replace(PH_IP, mServerIP)
		Case PLANID_DEVICES_ALL
			Url = URL_GET_DEVICES.Replace(PH_IP, mServerIP)
		Case PLANID_DEVICES_USED
			Url = URL_GET_DEVICES_USED.Replace(PH_IP, mServerIP)
		Case PLANID_DEVICES_NOTUSED
			Url = URL_GET_DEVICES_NOTUSED.Replace(PH_IP, mServerIP)
		Case PLANID_USERVARIABLES
			Url = URL_GET_USERVARIABLES.Replace(PH_IP, mServerIP)
		Case Else
			Url = URL_GET_PLAN_DEVICES.Replace(PH_IP, mServerIP).Replace(PH_IDX, PlanID)	
	End Select
	ToLog($"GetDevices: url=${Url}, planid=${PlanID}"$)
	'Clear the devices list
	Devices.Clear
	'Init the httpjob
	Job.Initialize("GetDevices", Me)
	Try
		Job.Download(Url)
		Wait For (Job) JobDone(Job As HttpJob)
		StatusSuccess = Job.Success
		If Job.Success Then
			'Get the list of devices with properties
			Devices = GetDevicesList(Job.GetString, PlanID)
			StatusMessage = MSG_STATUS_OK
		Else
			StatusMessage = $"[${MSG_STATUS_ERROR}] ${Job.ErrorMessage}"$
		End If
	Catch
		StatusSuccess = False
		StatusMessage = LastException.Message
	End Try
	Result = StatusSuccess
	ToLog($"GetDevices: laststatus=${StatusSuccess}, lastmessage=${StatusMessage}"$)
	Return Result
End Sub

'Get the status for a specific device.
'The class global Devicesis updated and can be used as required to get device(s) selected properties.
'Returns - Boolean True=OK, False=Error. Check class globals StatusSuccess and StatusMessage.
'Example: <code>
'Wait For(DomController.GetSpecificDevice(123)) Complete (Result As Boolean)
'xui.MsgboxAsync($"Status: ${DomController.StatusSuccess}, ${DomController.StatusMessage}, "$, "GetFavoriteDevices")
'If DomController.StatusSuccess Then
'	For Each Device As Map In DomController.DevicesList
'		clvDevices.AddTextItem(DomController.MapToString(Device), Device.Get(DomController.DOMOTICZ_KEY_IDX))
'	Next
'End If
'</code>
Public Sub GetSpecificDevice(Idx As String) As ResumableSub
	Dim Result As Boolean = False
	Dim Job As HttpJob
	Dim Url As String = URL_GET_DEVICE.Replace(PH_IP, mServerIP).Replace(PH_IDX, Idx)
	ToLog($"GetSpecificDevice: url=${Url}"$)
	'Clear the deviceslist
	Devices.Clear
	'Init the httpjob
	Job.Initialize("GetSpecificDevice", Me)
	Try
		Job.Download(Url)
		Wait For (Job) JobDone(Job As HttpJob)
		StatusSuccess = Job.Success
		If Job.Success Then
			'Select all devices from the devices list as these are already filtered by the selective device thru the URL
			Devices = GetDevicesList(Job.GetString, PLANID_DEVICES_ALL)
			StatusMessage = MSG_STATUS_OK
		Else
			StatusMessage = Job.ErrorMessage					
		End If
	Catch
		StatusSuccess = False
		StatusMessage = LastException.Message
	End Try
	Result = StatusSuccess
	ToLog($"GetSpecificDevice: laststatus=${StatusSuccess}, lastmessage=${StatusMessage}"$)
	Return Result
End Sub

'Get the list of devices from the json response string resulting from an JSON/API URL request.
'Each list entry holds the device properties.
'JsonResponse - Domoticz URL JSON/API request response string in JSON format
'PlanID - ID of the roomplan to select device. Set to -1 to select all devices, 0 for favortite devices else roomplan idx.
'Returns - List with all devices and selective properties.
'The selective properties are defined in class global mDeviceProperties.
Public Sub GetDevicesList(JsonResponse As String, PlanID As Int) As List
	Dim Result As List
	Dim Parser As JSONParser
	Dim ResponseRootMap As Map
	Dim DeviceList As List
	ToLog($"GetDevicesList: jsonresponselength=${JsonResponse.Length}, planid=${PlanID}"$)
	' Init the return list
	Result.Initialize
	' Check if the response has content. If not return an empty list (initialized)
	If JsonResponse.Trim.Length = 0 Then
		Return Result
	End If
	' Init the the JSON Response String Parser
	Parser.Initialize(JsonResponse)
	' Get the root of the JSON structure
	ResponseRootMap = Parser.NextObject
	' Get generic keys
	Dim app_version As String = ResponseRootMap.Get(DOMOTICZ_KEY_APPVERSION)
	Dim ServerTime As String = ResponseRootMap.Get(DOMOTICZ_KEY_SERVERTIME)
	' Get the result flag OK and status
	Dim title As String = ResponseRootMap.Get(DOMOTICZ_KEY_RETURN_TITLE)
	Dim status As String = ResponseRootMap.Get(DOMOTICZ_KEY_RETURN_STATUS)
	ToLog($"GetDevicesList: app_version=${app_version}, servertime=${ServerTime}, title=${title}, status=${status}"$)
	' List of devices from the key result"
	DeviceList = ResponseRootMap.Get("result")
	'Loop over each device entry and add to the devices list depending planid parameter
	For Each Device As Map In DeviceList
		Select PlanID
			Case PLANID_DEVICES_ALL
				Result.Add(Device)
			Case PLANID_DEVICES_USED
				Result.Add(Device)
			Case PLANID_DEVICES_NOTUSED
				Result.Add(Device)
			Case PLANID_USERVARIABLES
				'Replace special case map entries for the user variables
				'Get user variable key Value 
				Dim value As String = Device.Get(DOMOTICZ_KEY_VALUE)
				'Add new key Data with user variable value
				Device.Put(DOMOTICZ_KEY_DATA, value)
				'Update key SubType with UserVariable - do not change the Type as required for updating the user variable
				Device.Put(DOMOTICZ_KEY_SUBTYPE, DOMOTICZ_VALUE_SUBTYPE_USERVARIABLE)
				'Add key description without content - this field is used by clv item click to show the device information
				Device.Put(DOMOTICZ_KEY_DESCRIPTION, "")
				'Add
				Result.Add(Device)
			Case PLANID_DEVICES_FAVORITE
				If Device.Get(DOMOTICZ_KEY_FAVORITE).As(Int) == 1 Then
					Result.Add(Device)
				End If
			Case Else
				'RoomPlan Devices
				If Device.Get(DOMOTICZ_KEY_PLANID).As(Int) == PlanID Then
					Result.Add(Device)
				End If
		End Select
		ToLog($"GetDevicesList: ${MapToString(Device)}"$)
	Next
	'Sort by Name for the custom planids which have a value < 0
	If IIf(PlanID < 0, True, False) Then
		Result = GetDeviceListSorted(DEVICE_SORT_FIELDNAME, Result)
	End If
	'Return the (sorted) devices list with map entries
	ToLog($"GetDevicesList: done, size=${Result.Size}"$)
	Return Result
End Sub

'Sort the devices list by name, by using a list with sorttype - Type TDevice (Idx As Int, Name As String).
'The deviceslist has a map with all device properties given by Domoticz.
'A temp list is created with sorted entries by name using the type TDevice. This list is used to create a new sorted device list with map entries.
'DeviceList - List of devices with map entries
'Returns - Device list with map entries sorted by field Name
Public Sub GetDeviceListSorted(FieldName As String, DeviceList As List) As List
	'ToLog($"GetDeviceListSorted: devicelist=${DeviceList.size}"$)
	Dim Result As List
	'Create a temp list holding the a type device with name and idx sorted by name
	Dim DeviceListT As List
	DeviceListT.Initialize
	For Each Device As Map In DeviceList
		Dim DeviceT As TDevice
		DeviceT.Initialize
		DeviceT.Name = Device.Get(DOMOTICZ_KEY_NAME)
		DeviceT.Idx = Device.Get(DOMOTICZ_KEY_IDX).As(Int)
		DeviceListT.Add(DeviceT)
	Next
	'Sort the list by field Name
	If DeviceListT.Size > 0 Then DeviceListT.SortTypeCaseInsensitive(FieldName, True)
	'Using the sorted list with name & idx, create a new device sorted list holding the maps - the Result
	Result.Initialize
	For Each DeviceT As TDevice In DeviceListT
		'ToLog($"GetDeviceListSorted: name=${DeviceT.Name}, idx=${DeviceT.Idx}"$)
		Result.Add(GetDeviceFromIDX(DeviceT.Idx, DeviceList))
	Next
	ToLog($"GetDeviceListSorted: size=${DeviceListT.size}"$)
	Return Result
End Sub

'Get the list of device names only from the global var Devices.
'Returns - List of device names
Public Sub GetDeviceNamesList As List
	Dim Result As List
	Result.Initialize
	For Each Device As Map In Devices
		Result.Add(Device.Get(DOMOTICZ_KEY_NAME))
	Next
	Return Result
End Sub

'Get the device as map from its Idx. 
Public Sub GetDeviceFromIDX(Idx As Int, DeviceList As List) As Map
	ToLog($"GetDeviceFromIDX: idx=${Idx}"$)
	Dim Result As Map
	Result.Initialize
	For Each Device As Map In DeviceList
		If Device.Get(DOMOTICZ_KEY_IDX).As(Int) == Idx Then
			ToLog($"GetDeviceFromIDX: found=${Device.Get(DOMOTICZ_KEY_NAME)}"$)
			Result = Device
			Exit
		End If
	Next
	Return Result
End Sub

' Get the key value for a generic JSON object - not part of the property result (array)
' IMPORTANT: key must match exact!
' Response - Holding key:value pairs
' Key - Key (property) of the device
' Returns the value as a string or text "unknown" if the device or key was not found.
Public Sub GetGenericKey(Response As String, Key As String) As String
	Dim Result As String = "unknown"
	Dim Parser As JSONParser
	Dim RootObject As Map
	ToLog($"GetGenericKey: key=${Key}"$)
	' Parse the JSON Response String
	Parser.Initialize(Response)
	' Get the root of the JSON structure
	RootObject = Parser.NextObject
	' Get generic keys, i.e
	' ResponseRoot.Get("app_version")
	' ResponseRoot.Get("ServerTime")
	If (Response.Trim.Length > 0) Then
		ToLog($"GetGenericKey: found key=${RootObject.Get(Key)}"$)
		Return RootObject.Get(Key)
	End If
	Return Result
End Sub

' Get the value for a device key.
' IMPORTANT: key must match exact!
' Returns the value as a string or text "unknown" if the device or key was not found.
Public Sub GetDeviceKey(Response As String, Idx As Int, Key As String) As String
	Dim Result As String = "unknown"
	Dim Parser As JSONParser
	Dim RootObject As Map
	Dim ResultArray As List
	ToLog($"GetDeviceKey: idx=${Idx}, key=${Key}"$)
	' Parse the JSON Response String
	Parser.Initialize(Response)
	' Get the root of the JSON structure
	RootObject = Parser.NextObject
	' Check if response content
	If (Response.Trim.Length = 0) Then
		Return Result
	End If
	' Get the array of devices from the results key
	ResultArray = RootObject.Get("result")
	If ResultArray.IsInitialized Then
		' Get the key for a device idx
		' List of devices from the key result"
		For Each Device As Map In ResultArray
			Dim DevIdx As Int = Device.Get("idx")
			If DevIdx = Idx Then
				ToLog($"GetDeviceKey: found idx=${Idx}, key=${Key}, value=${Device.Get(Key)}"$)
				Return Device.Get(Key)
			End If
		Next
	End If
	Return Result
End Sub

' Get the device properties as mao from the result key. The result key holds an array of devices.
' Returns an initialised map if device found else an uninitialised map
Public Sub GetDeviceProperties(Response As String, Idx As Int) As Map
	Dim Result As Map
	Dim Parser As JSONParser
	Dim RootObject As Map
	Dim ResultArray As List

	ToLog($"GetDeviceProperties: idx=${Idx}"$)
	' Parse the JSON Response String
	Parser.Initialize(Response)
	' Get the root of the JSON structure
	RootObject = Parser.NextObject
	' Check if response content
	If (Response.Trim.Length = 0) Then
		Return Result
	End If
	' Get the array of devices from the results key
	ResultArray = RootObject.Get("result")
	If ResultArray.IsInitialized Then
		' Get the key for a device idx
		' List of devices from the key result"
		For Each Device As Map In ResultArray
			Dim DevIdx As Int = Device.Get("idx")
			If DevIdx = Idx Then
				' Return initialised map with device data
				Return Device
			End If
		Next
	End If
	' Return uninitialised map
	Return Result
End Sub

'Get a device type and subtype.
'Device - Map of all device properties
'Return - String with Type and SubType seperated by a comma
'Example: "Light/Switch,Switch" 
Public Sub GetDeviceTypeSubType(Device As Map) As String
	Dim Result As String = ""
	Result = $"${Device.Get(DOMOTICZ_KEY_TYPE)};${Device.Get(DOMOTICZ_KEY_SUBTYPE)}"$
	Return Result
End Sub

'Get the list of plans from the json response string resulting from an JSON/API URL request.
'The class global Plans is updated and can be used as required to get device(s) selected properties.
'Returns - Boolean True=OK, False=Error. Check class globals StatusSuccess and StatusMessage.
'Example: <code>
'Wait For(DomController.GetPlans) Complete (ResultState As Boolean)
'If DomController.StatusSuccess Then
'	Dim Plans As List
'	Plans.Initialize
'	For Each Plan As Map In DomController.Plans
'		Plans.Add(Plan.Get(DomController.DOMOTICZ_KEY_NAME))
'	Next
'End If
'</code>
Public Sub GetPlans As ResumableSub
	Dim Result As Boolean = False
	Dim Job As HttpJob
	Dim Url As String = URL_GET_PLANS.Replace(PH_IP, mServerIP)
	Dim Parser As JSONParser
	Dim ResponseRootMap As Map
	Dim PlanList As List
	ToLog($"GetPlans: url=${Url}"$)
	'Clear the global list Rooms
	Plans.Clear
	'Init the httpjob
	Job.Initialize("GetPlans", Me)
	Try
		Job.Download(Url)
		Wait For (Job) JobDone(Job As HttpJob)
		StatusSuccess = Job.Success
		If Job.Success Then
			' Check if the response has content. If not return an empty list (initialized)
			If Job.GetString.Trim.Length = 0 Then
				Return Result
			End If
			' Init the the JSON Response String Parser
			Parser.Initialize(Job.GetString)
			' Get the root of the JSON structure
			ResponseRootMap = Parser.NextObject
			' List of plans from the key result"
			PlanList = ResponseRootMap.Get("result")
			'Loop over each plan entry and add to the global list Plans
			'Prior doing so, custom plan are added
			Plans.Add(JSONToMap(PLAN_DEVICES_FAVORITES))
			Plans.Add(JSONToMap(PLAN_DEVICES_ALL))
			Plans.Add(JSONToMap(PLAN_DEVICES_USED))
			Plans.Add(JSONToMap(PLAN_DEVICES_NOTUSED))
			Plans.Add(JSONToMap(PLAN_USERVARIABLES))
			Plans.Add(JSONToMap(PLAN_ROOMS_SEPERATOR))		'Indicator to show list of rooms starts
			For Each Plan As Map In PlanList
				Plans.Add(Plan)
			Next
			ToLog($"GetPlans: plans=${MapToString(Plan)}"$)
		End If
	Catch
		StatusSuccess = False
		StatusMessage = LastException.Message
	End Try
	Result = StatusSuccess
	ToLog($"GetPlans: laststatus=${StatusSuccess}, lastmessage=${StatusMessage}"$)
	Return Result
End Sub

'Get the status for the devices assigned to a plan.
'The class global Devicesis updated and can be used as required to get device(s) selected properties.
'Returns - Boolean True=OK, False=Error. Check class globals StatusSuccess and StatusMessage.
'</code>
Public Sub GetPlanDevices(PlanID As String) As ResumableSub
	Dim Result As Boolean = False
	Dim Job As HttpJob
	Dim Url As String = URL_GET_PLAN_DEVICES.Replace(PH_IP, mServerIP).Replace(PH_IDX, PlanID)
	ToLog($"GetPlanDevices: url=${Url}, planid=${PlanID}"$)
	'Clear the devices list
	Devices.Clear
	'Init the httpjob
	Job.Initialize("GetPlanDevices", Me)
	Try
		Job.Download(Url)
		Wait For (Job) JobDone(Job As HttpJob)
		StatusSuccess = Job.Success
		If Job.Success Then
			Devices = GetDevicesList(Job.GetString, PLANID_DEVICES_ALL)
			StatusMessage = MSG_STATUS_OK
		Else
			StatusMessage = $"[${MSG_STATUS_ERROR}] ${Job.ErrorMessage}"$
		End If
	Catch
		StatusSuccess = False
		StatusMessage = LastException.Message
	End Try
	Result = StatusSuccess
	ToLog($"GetPlanDevices: laststatus=${StatusSuccess}, lastmessage=${StatusMessage}"$)
	Return Result
End Sub

'Get the list of user variables.
'The class global UserVariables updated and can be used as required to get the uservariable selected properties.
'Returns - Boolean True=OK, False=Error. Check class globals StatusSuccess and StatusMessage.
Public Sub GetUserVariables As ResumableSub
	Dim Result As Boolean = False
	Dim Job As HttpJob
	Dim Url As String = URL_GET_USERVARIABLES.Replace(PH_IP, mServerIP)
	ToLog($"GetUserVariables: url=${Url}"$)
	'Clear the devices list
	UserVariables.Clear
	'Init the httpjob
	Job.Initialize("GetUserVariables", Me)
	Try
		Job.Download(Url)
		Wait For (Job) JobDone(Job As HttpJob)
		StatusSuccess = Job.Success
		If Job.Success Then
			Dim Data As String = Job.GetString
			Data = Data.Replace(DOMOTICZ_KEY_VALUE, DOMOTICZ_KEY_DATA)
			Devices = GetDevicesList(Data, PLANID_DEVICES_ALL)
			StatusMessage = MSG_STATUS_OK
		Else
			StatusMessage = $"[${MSG_STATUS_ERROR}] ${Job.ErrorMessage}"$
		End If
	Catch
		StatusSuccess = False
		StatusMessage = LastException.Message
	End Try
	Result = StatusSuccess
	ToLog($"GetUserVariables: laststatus=${StatusSuccess}, lastmessage=${StatusMessage}"$)
	Return Result
End Sub

'Get the type of a uservariable. Defined as global array mUserVariabelTypes.
'0 = Integer, e.g. -1, 1, 0,
'1 = Float, e.g. -1.1, 1.2, 3.1
'2 = String
'3 = Date in format DD/MM/YYYY
'4 = Time in 24 hr format HH:MM
'UserVariableType - Integer
'Returns - String with the type of the user variable
Public Sub GetUserVariableType(UserVariableType As Int) As String
	Dim Result As String = "Unknown"
	If UserVariableType >= 0 And UserVariableType <= 4 Then
		Result = mUserVariabelTypes(UserVariableType)
	End If
	ToLog($"GetUserVariableType: type=${Result}"$)
	Return Result
End Sub

'Get the full url of the log of a Domoticz device.
'Idx - Idx of the device
'Return - Full URL to show the log in a browser
Public Sub GetDeviceLogUrl(Idx As String) As String
	Return URL_LOG_DEVICE.Replace(PH_IP, mServerIP).Replace(PH_IDX, Idx)
End Sub

#End Region

#Region SETTER

'Set the Domoticz server ip global var.
'Value - Domoticz server ip.
Public Sub setServerIP(Value As String)
	mServerIP = Value
	ToLog($": serverip=${Value}, isurl=${IsUrl(Value)}"$)
'	If IsUrl(Value) Then mServerIP = Value
End Sub

'Get the Domoticz server ip.
'Returns - Domoticz server IP.
Public Sub getServerIP As String
	Return mServerIP
End Sub

'Set the state of a switch to On, Off, Toggle, Stop.
'Idx - Idx of the device
'SwitchCmd - On, Off, Toggle, Stop
'Returns - Boolean True=OK, False=Error. Check class globals StatusSuccess and StatusMessage.
'Example: <code>
'Wait For(DomController.Dialog_SwitchLight(deviceDialog, Title, Label, Value)) Complete (SwitchLightResult As Boolean)
'If DomController.StatusSuccess Then
'	'Set the switch depending ResultValue True=On, or False=Off
'	Dim SwitchCmd As String = IIf(SwitchLightResult, "On", "Off")
'	Wait For(DomController.SetSwitchLight(device.Get(DomController.DOMOTICZ_KEY_IDX), SwitchCmd)) Complete (ResultValue As Boolean)
'End If
'</code>
Public Sub SetSwitchLight(Idx As String, SwitchCmd As String) As ResumableSub
	Dim Result As Boolean = False
	Dim Job As HttpJob
	Dim Url As String = URL_SWITCHLIGHT.Replace(PH_IP, mServerIP).Replace(PH_IDX, Idx).Replace(PH_SWITCHCMD, SwitchCmd)
	ToLog($"SetSwitchLight: url=${Url}, switchcmd=${SwitchCmd}"$)
	'Init the httpjob
	Job.Initialize("SetSwitchLight", Me)
	Try
		Job.Download(Url)
		Wait For (Job) JobDone(Job As HttpJob)
		StatusSuccess = Job.Success
		If Job.Success Then
			StatusMessage = MSG_STATUS_OK
		Else
			StatusMessage = Job.ErrorMessage
		End If
	Catch
		StatusSuccess = False
		StatusMessage = LastException.Message
	End Try
	Result = StatusSuccess
	ToLog($"SetSwitchLight: laststatus=${StatusSuccess}, lastmessage=${StatusMessage}"$)
	Return Result
End Sub

'Set the level of a selector switch.
'Idx - Idx of the device
'Level - Level of the selector as defined in the selectors edit page (0 = Off)
'Returns - Boolean True=OK, False=Error. Check class globals StatusSuccess and StatusMessage.
'Example: <code>
'Wait For(DomController.Dialog_SwitchLight(deviceDialog, Title, Label, Value)) Complete (SwitchLightResult As Boolean)
'If DomController.StatusSuccess Then
'	'Set the switch depending ResultValue True=On, or False=Off
'	Dim SwitchCmd As String = IIf(SwitchLightResult, "On", "Off")
'	Wait For(DomController.SetSwitch(device.Get(DomController.DOMOTICZ_KEY_IDX), SwitchCmd)) Complete (ResultValue As Boolean)
'End If
'</code>
Public Sub SetSwitchLightSetLevel(Idx As String, Level As String) As ResumableSub
	Dim Result As Boolean = False
	Dim Job As HttpJob
	Dim Url As String = URL_SWITCHLIGHT_SETLEVEL.Replace(PH_IP, mServerIP).Replace(PH_IDX, Idx).Replace(PH_LEVEL, Level)
	ToLog($"SetSwitchLightSetLevel: url=${Url}, level=${Level}"$)
	'Init the httpjob
	Job.Initialize("SetSwitchLightSetLevel", Me)
	Try
		Job.Download(Url)
		Wait For (Job) JobDone(Job As HttpJob)
		StatusSuccess = Job.Success
		If Job.Success Then
			StatusMessage = MSG_STATUS_OK
		Else
			StatusMessage = Job.ErrorMessage
		End If
	Catch
		StatusSuccess = False
		StatusMessage = LastException.Message
	End Try
	Result = StatusSuccess
	ToLog($"SetSwitchLightSetLevel: laststatus=${StatusSuccess}, lastmessage=${StatusMessage}"$)
	Return Result
End Sub

'Set the new thermostat setpoint
'Returns - Boolean True=OK, False=Error. Check class globals StatusSuccess and StatusMessage.
'Example: <code>
'Wait For(DomController.Dialog_Positive(deviceDialog, Title, Label, Value)) Complete (SetSetpointResult As Float)
'If DomController.StatusSuccess Then
'	Wait For(DomController.SetSetpoint(device.Get(DomController.DOMOTICZ_KEY_IDX), SetSetpointResult)) Complete (ResultValue As Boolean)
'End If
'</code>
Public Sub SetSetPoint(Idx As String, Value As Float) As ResumableSub
	Dim Result As Boolean = False
	Dim Job As HttpJob
	Dim Url As String = URL_SETSETPOINT.Replace(PH_IP, mServerIP).Replace(PH_IDX, Idx).Replace(PH_SETPOINT, Value)
	ToLog($"SetSetpoint: url=${Url}"$)
	'Init the httpjob
	Job.Initialize("SetSetpoint", Me)
	Try
		Job.Download(Url)
		Wait For (Job) JobDone(Job As HttpJob)
		StatusSuccess = Job.Success
		If Job.Success Then
			StatusMessage = MSG_STATUS_OK
		Else
			StatusMessage = Job.ErrorMessage
		End If
	Catch
		StatusSuccess = False
		StatusMessage = LastException.Message
	End Try
	Result = StatusSuccess
	ToLog($"SetSetpoint: laststatus=${StatusSuccess}, lastmessage=${StatusMessage}"$)
	Return Result
End Sub

'Set a new value for a user variable
'Returns - Boolean True=OK, False=Error. Check class globals StatusSuccess and StatusMessage.
'Example: <code>
'Wait For(DomController.Dialog_Text(deviceDialog, Title, Label, Value)) Complete (UserVariableResult As String)
'If DomController.StatusSuccess Then				
'	Wait For(DomController.SetUserVariable(UserVariableName, UserVariableType, UserVariableResult)) Complete (ResultValue As Boolean)
'End If			
'</code>
Public Sub SetUserVariable(UserVariableName As String, UserVariableType As String, UserVariableValue As String) As ResumableSub
	Dim Result As Boolean = False
	Dim Job As HttpJob
	Dim Url As String = URL_SET_USERVARIABLE.Replace(PH_IP, mServerIP).Replace(PH_USERVARIABLENAME, UserVariableName).Replace(PH_USERVARIABLETYPE, UserVariableType).Replace(PH_USERVARIABLEVALUE, UserVariableValue)
	ToLog($"SetUserVariable: url=${Url}, name=${UserVariableName}, type=${UserVariableType}, value=${UserVariableValue}"$)
	'Init the httpjob
	Job.Initialize("SetUserVariable", Me)
	Try
		Job.Download(Url)
		Wait For (Job) JobDone(Job As HttpJob)
		StatusSuccess = Job.Success
		If Job.Success Then
			StatusMessage = MSG_STATUS_OK
		Else
			StatusMessage = Job.ErrorMessage
		End If
	Catch
		StatusSuccess = False
		StatusMessage = LastException.Message
	End Try
	Result = StatusSuccess
	ToLog($"SetUserVariable: laststatus=${StatusSuccess}, lastmessage=${StatusMessage}"$)
	Return Result
End Sub
#End Region

#Region DEVICESMAPPING
'https://www.domoticz.com/wiki/Developing_a_Python_plugin#Available_Device_Types
'https://www.domoticz.com/wiki/Domoticz_API/JSON_URL%27s

'Create the CSV list (with ; seperator) of device groups used by for example device group icon weather.
'The CSV string contains: Name;Type;SubType;FontAwesomeIcon
'Returns - List
Public Sub SetDeviceGroups As List
	Dim DevGroups As List
	DevGroups.Initialize
	DevGroups.Add($"${GROUP_WEATHER};Temp;LaCrosse TX3;${Chr(0xF2C8)}"$)											'Thermometer
	DevGroups.Add($"${GROUP_WEATHER};Temp + Humidity;Cresta; TFA TS34C;${GROUP_ICON_WEATHER}"$)
	DevGroups.Add($"${GROUP_WEATHER};Temp + Hum + Baro;THB1 - BTHR918, BTHGN129;${GROUP_ICON_WEATHER}"$)
	DevGroups.Add($"${GROUP_WEATHER};Temp + Hum + Baro;THB2 - BTHR918N, BTHR968;${GROUP_ICON_WEATHER}"$)
	DevGroups.Add($"${GROUP_WEATHER};Temp + Humidity;THGN122/123/132, THGR122/228/238/268;${GROUP_ICON_WEATHER}"$)
	DevGroups.Add($"${GROUP_WEATHER};Temp + Humidity;Cresta, TFA TS34C;${GROUP_ICON_WEATHER}"$)
	DevGroups.Add($"${GROUP_WEATHER};Temp + Hum + Baro;Weather Station;${GROUP_ICON_WEATHER}"$)
	DevGroups.Add($"${GROUP_WEATHER};Wind;WTGR800;${GROUP_ICON_WEATHER}"$)											
	DevGroups.Add($"${GROUP_WEATHER};Wind;TFA;${GROUP_ICON_WEATHER}"$)
	DevGroups.Add($"${GROUP_WEATHER};Wind;Weather Station;${GROUP_ICON_WEATHER}"$)
	DevGroups.Add($"${GROUP_WEATHER};General;Visibility;${GROUP_ICON_WEATHER}"$)
	DevGroups.Add($"${GROUP_WEATHER};General;Barometer;${GROUP_ICON_WEATHER}"$)
	DevGroups.Add($"${GROUP_WEATHER};Lux;Lux;${GROUP_ICON_WEATHER}"$)
	DevGroups.Add($"${GROUP_WEATHER};UV;UVN128;UV138;${GROUP_ICON_WEATHER}"$)

	DevGroups.Add($"${GROUP_SWITCHES};Color Switch;WW;${GROUP_ICON_SWITCHES}"$)
	DevGroups.Add($"${GROUP_SWITCHES};Color Switch;RGB;${GROUP_ICON_SWITCHES}"$)
	DevGroups.Add($"${GROUP_SWITCHES};Color Switch;RGBWW;${GROUP_ICON_SWITCHES}"$)
	DevGroups.Add($"${GROUP_SWITCHES};Light/Switch;Switch;${Chr(0xF205)}"$)						'Switch icon
	DevGroups.Add($"${GROUP_SWITCHES};Light/Switch;Selector Switch;${GROUP_ICON_SWITCHES}"$)	'Switch icon
	DevGroups.Add($"${GROUP_SWITCHES};Lighting 1;Impuls;${GROUP_ICON_SWITCHES}"$)
	DevGroups.Add($"${GROUP_SWITCHES};Lighting 1;ELRO AB400;${GROUP_ICON_SWITCHES}"$)
	DevGroups.Add($"${GROUP_SWITCHES};Lighting 2;AC;${GROUP_ICON_SWITCHES}"$)
	DevGroups.Add($"${GROUP_SWITCHES};Lighting 4;PT2262;${GROUP_ICON_SWITCHES}"$)
	DevGroups.Add($"${GROUP_SWITCHES};RFY;RFY;${GROUP_ICON_SWITCHES}"$)

	DevGroups.Add($"${GROUP_UTILITY};Thermostat;SetPoint;${Chr(0xF2CB)}"$)
	DevGroups.Add($"${GROUP_UTILITY};General;kWh;${GROUP_ICON_UTILITY}"$)
	DevGroups.Add($"${GROUP_UTILITY};Usage;Electric;${GROUP_ICON_UTILITY}"$)

	DevGroups.Add($"${GROUP_INFORMATION};General;Text;${Chr(0xF0F6)}"$)							'Text
	DevGroups.Add($"${GROUP_INFORMATION};General;Alert;${Chr(0xF071)}"$)						'Alert icon
	DevGroups.Add($"${GROUP_INFORMATION};General;Distance;${GROUP_ICON_INFORMATION}"$)
	DevGroups.Add($"${GROUP_INFORMATION};General;Soil Moisture;${GROUP_ICON_INFORMATION}"$)
	DevGroups.Add($"${GROUP_INFORMATION};General;Percentage;${Chr(0xF295)}"$)					'Percentage
	DevGroups.Add($"${GROUP_INFORMATION};General;Custom Sensor;${Chr(0xF1BD)}"$)

	DevGroups.Add($"${GROUP_SECURITY};Security;X10 security;${GROUP_ICON_SECURITY}"$)
	Return DevGroups
End Sub

'Get the device group for a device type and subtype
'Example device group information for a device type General and subtype text: information,General,Text
Public Sub GetDeviceGroup(DevType As String, DevSubType As String) As String
	For Each Entry As String In mDeviceGroups
		Dim Items() As String = Regex.Split(";", Entry)
		'Example: information,General,Text for Items(0),Items(1),Items(3)
		If Items.Length == 3 Then
			If Items(1) == DevType And Items(2) == DevSubType Then
				Return Items(0)
			End If
		End If
	Next
	Return ""
End Sub

'Get the device Group fontawesome icon for a device type and subtype
Public Sub GetDeviceGroupIcon(DevType As String, DevSubType As String) As String
	For Each Entry As String In mDeviceGroups
		Dim Items() As String = Regex.Split(";", Entry)
		'Example CSV string to Items:
		'Information;General;Text;Chr(0xF1EA) for Items(0),Items(1),Items(2),Items(3)
		If Items(1) == DevType And Items(2) == DevSubType Then Return Items(3)
	Next
	Return GROUP_ICON_UNKNOWN
End Sub
#End Region

#Region DIALOGS
'Definition of the various dialogs to set data or action.
'TODO: Select a dialog depending device type and subtype.
'Examples:
'if type="General" and SubType="Alert" then open Dialog_Information
'if type="Light/Switch" and SubType="Switch" then open Dialog_Switch

'Get the dialog id to open. In case a device value can be changed a b4xdialog is opened else xui.msgdialog.
'See https://www.domoticz.com/wiki/Developing_a_Python_plugin#Available_Device_Types
Public Sub GetDialogID(Device As Map) As Int
	Dim Result As Int = DIALOGID_UNKNOWN
	'Get the device type and subtype to select the dialog
	Select GetDeviceTypeSubType(Device)
		Case "Light/Switch;Switch"
			Result = DIALOGID_SWITCHLIGHT
		Case "Light/Switch;Selector Switch"
			Result = DIALOGID_SWITCHLIGHT_SETLEVEL
		Case "Thermostat;SetPoint"
			Result = DIALOGID_SETSETPOINT
		'Add more
		Case Else
			Result = DialogID_INFORMATION
	End Select
	'Special case uservariable
	If Device.Get(DOMOTICZ_KEY_SUBTYPE) == DOMOTICZ_VALUE_SUBTYPE_USERVARIABLE Then
		Result = DIALOGID_USERVARIABLE
	End If
	Return Result
End Sub

'Dialog for a whole positive or negative number.
'Dialog - Global dialog object
'Title - Dialog title
'Value - Predefined input value as Int
'Returns - Result value as Int
'The global class flag StatusSuccess indicates if the dialog completed as OK (true) or Cancel (false).
Public Sub Dialog_Whole(Dialog As B4XDialog, Title As String, Label As String, Value As Int) As ResumableSub
	Dim ResultValue As Int
	Dim input As B4XInputTemplate
	StatusSuccess = False
	Dialog.Title = Title
	input.Initialize
	input.lblTitle.Text = Label
	input.Text = Value
	input.ConfigureForNumbers(False, True)
	Wait For (Dialog.ShowTemplate(input, "OK", "", "CANCEL")) Complete (Result As Int)
	If Result = xui.DialogResponse_Positive Then
		StatusSuccess = True
		ResultValue = input.Text 'no need to check with IsNumber
		'Dialog.Show($"2 * $1.2{res} = $1.2{2 * res}"$, "OK", "", "")
	End If
	Return ResultValue
End Sub

'Dialog for a whole positive number.
'Dialog - Global dialog object
'Title - Dialog title
'Value - Predefined input value as Int
'Returns - Result value as Int
'The global class flag StatusSuccess indicates if the dialog completed as OK (true) or Cancel (false).
Public Sub Dialog_WholePositive(Dialog As B4XDialog, Title As String, Label As String, Value As Int) As ResumableSub
	Dim ResultValue As Int
	Dim Input As B4XInputTemplate
	StatusSuccess = False
	Dialog.Title = Title
	Input.Initialize
	Input.lblTitle.Text = Label
	Input.Text = Value
	Input.ConfigureForNumbers(False, False)
	Wait For (Dialog.ShowTemplate(Input, "OK", "", "CANCEL")) Complete (Result As Int)
	If Result = xui.DialogResponse_Positive Then
		StatusSuccess = True
		ResultValue = Input.Text 'no need to check with IsNumber
		' Dialog.Show($"2 * $1.2{Result} = $1.2{2 * Result}"$, "OK", "", "")
	End If
	Return ResultValue
End Sub

'Dialog for a positive number.
'Dialog - Global dialog object
'Title - Dialog title
'Value - Predefined input value as Float
'Returns - Result value as Float
'The global class flag StatusSuccess indicates if the dialog completed as OK (true) or Cancel (false).
Public Sub Dialog_Positive(Dialog As B4XDialog, Title As String, Label As String, Value As Float) As ResumableSub
	Dim ResultValue As Float
	Dim Input As B4XInputTemplate
	StatusSuccess = False
	Dialog.Title = Title
	Input.Initialize
	Input.lblTitle.Text = Label
	Input.Text = Value
	Input.ConfigureForNumbers(True, False)
	Wait For (Dialog.ShowTemplate(Input, "OK", "", "CANCEL")) Complete (Result As Int)
	If Result = xui.DialogResponse_Positive Then
		StatusSuccess = True
		ResultValue = Input.Text 'no need to check with IsNumber
		' Dialog.Show($"2 * $1.2{res} = $1.2{2 * res}"$, "OK", "", "")
	End If
	Return ResultValue
End Sub

'Dialog for any number.
'Dialog - Global dialog object
'Title - Dialog title
'Value - Predefined input value as Float
'Returns - Result value as Float
'The global class flag StatusSuccess indicates if the dialog completed as OK (true) or Cancel (false).
Public Sub Dialog_Number(Dialog As B4XDialog, Title As String, Label As String, Value As Float) As ResumableSub
	Dim Input As B4XInputTemplate
	Dim ResultValue As Float
	StatusSuccess = False
	Dialog.Title = Title
	Input.Initialize
	Input.lblTitle.Text = Label
	Input.Text = Value
	Input.ConfigureForNumbers(True, True)
	Wait For (Dialog.ShowTemplate(Input, "OK", "", "CANCEL")) Complete (Result As Int)
	If Result = xui.DialogResponse_Positive Then
		StatusSuccess = True
		ResultValue = Input.Text
		'Dialog.Show($"2 * $1.2{res} = $1.2{2 * res}"$, "OK", "", "")
	End If
	Return ResultValue
End Sub

'Dialog for any input i.e. text, numbers, email address etc..
'Dialog - Global dialog object
'Title - Dialog title
'Value - Predefined input value as Text
'Returns - Result value as Text
'The global class flag StatusSuccess indicates if the dialog completed as OK (true) or Cancel (false).
Public Sub Dialog_Text(Dialog As B4XDialog, Title As String, Label As String, Value As String) As ResumableSub
	Dim Input As B4XInputTemplate
	Dim ResultValue As String
	StatusSuccess = False
	Dialog.Title = Title
	Input.Initialize
	Input.lblTitle.Text = Label
	Input.Text = Value
	Input.mBase.Width = 500dip
	Wait For (Dialog.ShowTemplate(Input, "OK", "", "CANCEL")) Complete (Result As Int)
	If Result = xui.DialogResponse_Positive Then
		StatusSuccess = True
		ResultValue = Input.Text
	End If
	Return ResultValue
End Sub

'Free form dialog.
'Dialog - Global dialog object
'Title - Dialog title
'Value - Predefined input value as string
'Returns - Result value as string
'The global class flag StatusSuccess indicates if the dialog completed as OK (true) or Cancel (false).
Public Sub Dialog_FreeForm(Dialog As B4XDialog, Title As String, Label As String, Value As String) As ResumableSub
	Dim Input As B4XInputTemplate
	Dim ResultValue As String
	StatusSuccess = False
	Dialog.Title = Title
	Input.Initialize
	Input.lblTitle.Text = Label
	Input.Text = Value
	Input.RegexPattern = ".+" 'require at least one character
	Wait For (Dialog.ShowTemplate(Input, "OK", "", "CANCEL")) Complete (Result As Int)
	If Result = xui.DialogResponse_Positive Then
		StatusSuccess = True
		ResultValue = Input.Text
		'Dialog.Show(Input.Text, "OK", "", "")
	End If
	Return ResultValue
End Sub

'Information dialog using free form with read only input field.
'Dialog - Global dialog object
'Title - Dialog title
'Value - Predefined input value as string
'Returns - Nothing
'The global class flag StatusSuccess indicates if the dialog completed as OK (true) or Cancel (false).
Public Sub Dialog_Information(Dialog As B4XDialog, Title As String, Label As String, Value As String) As ResumableSub
	Dim Input As B4XInputTemplate
	Dim ResultValue As Boolean
	StatusSuccess = True
	ResultValue = StatusSuccess
	Dialog.Title = Title
	Input.Initialize
	Input.lblTitle.Text = Value
	Input.Text = Value
	Input.TextField1.Visible = False
	Wait For (Dialog.ShowTemplate(Input, "OK", "", "")) Complete (Result As Int)
	Return ResultValue
End Sub

'Dialog to set a switch to ON or OFF.
'Dialog - Global dialog object
'Title - Dialog title
'Value - Predefined input value as string: On or Off
'Returns - Result value as boolean True=On or False=Off
'The global class flag StatusSuccess indicates if the dialog completed as OK (true) or Cancel (false).
Public Sub Dialog_SwitchLight(Dialog As B4XDialog, Title As String, Label As String, Value As String) As ResumableSub
	Dim ResultValue As Boolean
	Dim Input As B4XInputTemplate
	StatusSuccess = True
	Dialog.Title = Title
	Input.Initialize
	Input.lblTitle.Text = Label
	Input.Text = Value
	Input.TextField1.Visible = False
	'The dialog buttons YES/NO are used to set the switch to On (YES, positive) or Off (NO, negative) or Cancel (Do nothing)
	Wait For (Dialog.ShowTemplate(Input, "ON", "OFF", "Cancel")) Complete (Result As Int)
	'Handle button response
	If Result = xui.DialogResponse_Positive Then ResultValue = True		'On
	If Result = xui.DialogResponse_Negative Then ResultValue = False	'Off
	If Result = xui.DialogResponse_Cancel Then 							'Cancel
		ResultValue = False
		StatusSuccess = False
	End If
	Return ResultValue
End Sub

'Dialog to set a switch level (o=Off).
'Dialog - Global dialog object
'Title - Dialog title
'Value - Predefined input value as string: On or Off
'Returns - Result value as boolean True=On or False=Off
'The global class flag StatusSuccess indicates if the dialog completed as OK (true) or Cancel (false).
Public Sub Dialog_SwitchLightSetLevel(Dialog As B4XDialog, Title As String, LevelNames As String, Value As Int) As ResumableSub
	Dim ResultValue As Int = -1
	Dim SearchTemplate As B4XSearchTemplate
	Dim LevelNamesList As List
	LevelNamesList.Initialize
	LevelNamesList.AddAll(GetLevelNames(LevelNames))
	ToLog($"Dialog_SwitchLightSetLevel: levelnameslist=${LevelNamesList}, selecteditem=${LevelNamesList.Get(Value * 0.1)}"$)
	SearchTemplate.Initialize
	SearchTemplate.SetItems(LevelNamesList)
	SearchTemplate.SearchField.TextField.Visible = False
	StatusSuccess = False
	Dialog.Title = Title
	Dim rs As Object = Dialog.ShowTemplate(SearchTemplate, "OK", "", "Cancel")
	Sleep(100)
	' SearchTemplate.SearchField.Text = "Information"	' LevelNamesList.Get(ItemIndex)
	Wait For (rs) Complete (Result As Int)
	ToLog($"Dialog_SwitchLightSetLevel: result=${Result}, selecteditem=${SearchTemplate.SelectedItem}"$)
	If Result = xui.DialogResponse_Positive Then
		StatusSuccess = True
		Dim Index As Int = LevelNamesList.IndexOf(SearchTemplate.SelectedItem)
		ResultValue = Index * 10
		ToLog($"Dialog_SwitchLightSetLevel: selecteditem=${SearchTemplate.SelectedItem}, index=${Index}"$)
	End If
	Return ResultValue
End Sub

'Sub btnCustom_Click
'	Dim p As B4XView = xui.CreatePanel("")
'	p.SetLayoutAnimated(0, 0, 0, 300dip, 150dip)
'	p.LoadLayout("CustomDialog")
'	dialog.PutAtTop = True 'put the dialog at the top of the screen
'	Dim rs As ResumableSub = dialog.ShowCustom(p, "OK", "", "CANCEL")
'	fieldFirstName.TextField.RequestFocus
'	Wait For (rs) Complete (Result As Int)
'	If Result = xui.DialogResponse_Positive Then
'		dialog.Show(fieldFirstName.Text & " " & fieldLastName.Text, "OK", "", "")
'	End If
'End Sub
#End Region

#Region HELPERS
'Write a message to the log with time as prefix text.
Private Sub ToLog(msg As String)
	If mDebug Then
		Log($"${DateTime.Time(DateTime.Now)} ${msg}"$)
	End If
End Sub

'Export a map to csv string.
'm - Initialized map.
Public Sub MapToString(m As Map) As String
	Dim Result As String
	If m.IsInitialized Then
		For Each Key As String In m.Keys
			Dim Value As String = m.Get(Key)
			Result = IIf(Result == "", $"${Key}=${Value}"$, $"${Result}, ${Key}=${Value}"$)
		Next
	End If
	Return Result
End Sub

'Converts a JSON string to a Map.
'value - JSON string
'Returns - Map with the JSON structure
Public Sub JSONToMap(value As String) As Map
	Dim JSON As JSONParser
	JSON.Initialize(value)
	Return JSON.NextObject
End Sub

'Source (Many thanks for providing): https://stackoverflow.com/a/3809435/971547
Public Sub IsUrl (s As String) As Boolean	'ignore
	Return Regex.IsMatch("https?:\/\/(www\.)?[-a-zA-Z0-9@:%._\+~#=]{1,256}\.[a-zA-Z0-9()]{1,6}\b([-a-zA-Z0-9()@:%_\+.~#?&//=]*)", s)
End Sub

'Source (Many thanks for providing): https://www.b4x.com/android/forum/threads/xui-sd-base64-encode-decode.105755/
Public Sub Base64Decode(S As String) As String
	Dim base64chars As String = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/"
	Dim P,R = "" As String

    #IF B4J or B4A
	Dim Ob As JavaObject = S 'ignore
	S=Ob.RunMethod("replaceAll",Array As Object("[^ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=]",""))
    #ELSE
	' Don't known - Working Progress
    #End If

	If S.CharAt(S.Length-1)="=" Then
		If S.CharAt(S.Length-2)="=" Then
			p="AA"
		Else
			P="A"
		End If
	Else
		P=""
	End If
	S=S.SubString2(0,S.Length-P.Length) & P

	For C = 0 To S.Length-1 Step 4
		Try
			Dim n As Int = Bit.ShiftLeft(base64chars.IndexOf(S.CharAt(C)),18) + Bit.ShiftLeft(base64chars.IndexOf(S.CharAt(C+1)),12) + Bit.ShiftLeft(base64chars.IndexOf(S.CharAt(C+2)),6)+ base64chars.IndexOf(S.CharAt(C+3))
			r = r & Chr(Bit.And(Bit.ShiftRight(n,16),0xFF)) & Chr(Bit.And(Bit.ShiftRight(n,8),0xFF)) & Chr(Bit.And(n,0xFF))
		Catch
			Log("Position: " & C)
		End Try
	Next

	Return R.SubString2(0, R.Length - P.Length)
End Sub

'Get the selector switch level names as list.
'LevelNames - Device property LevelNames. Needs to be Base64Decoded.
Public Sub GetLevelNames(LevelNames As String) As List
	Dim Result As List
	Result.Initialize
	'Base64 Decode the levelnames
	LevelNames = Base64Decode(LevelNames)
	LevelNames = LevelNames.Replace("|", ";")
	'Split and add the level names to the list
	Result.AddAll(Regex.Split(";", LevelNames))
	ToLog($"GetLevelNames: result=${Result}"$)
	Return Result
End Sub

'Get the selector level name from its level index.
'Index - Level index 0, 10, 20 ...
'LevelNames - Device property LevelNames. Needs to be Base64Decoded
Public Sub GetLevelName(Index As Int, LevelNames As String) As String
	Dim Result As String
	Dim LevelNamesList As List
	Index = (Index * 0.1).As(Int)
	LevelNamesList = GetLevelNames(LevelNames)
	ToLog($"GetLevelName: Index=${Index} (${Index}), levelnames=${LevelNames}, size=${LevelNamesList.size}"$)
	If LevelNamesList.Size > 0  And Index <= LevelNamesList.Size - 1 Then
		Result = LevelNamesList.Get(Index)	
	End If
	Return Result
End Sub

#End Region
